
# Copyright 2022, Frank Sauerburger

"""Implementation of a q-gram index"""

import collections

class QGramIndex:
    """Q-gram index for fuzzy lookup"""

    def __init__(self, q_param=3):
        self.q_param = q_param
        self._index = collections.defaultdict(set)


    def add_term(self, term):
        """Add a single term to the index """
        qgrams = self._chunk(term)
        for qgram in qgrams:
            self._index[qgram].add(term)


    def search(self, query):
        """Find vocabulary terms matching the query"""
        qgrams = self._chunk(query)
        n_query = len(qgrams)

        matches = []
        for qgram in qgrams:
            matches.extend(self._index[qgram])

        counter = collections.Counter(matches)

        # Filter matches by number of qgram overlap
        # See Jaccard coefficient:
        # *Introduction to Information Retrieval*
        result = []
        for term, count in counter.items():
            n_term = self._n_qgram(term)
            jaccard = count / (n_term + n_query - count)
            result.append((jaccard, term))

        result.sort(key=lambda x: x[0])

        return result

    def _n_qgram(self, term):
        """Return the number of qgrams"""
        return max(1, len(term) + 3 - self.q_param)

    def _chunk(self, term):
        """
        Split an input term into a list of qgrams

        The beginning and end of the string is denoted by a dollar sign ($).
        If the term is too short, the qgrams might be shorter than q.
        """
        term = f"${term}$"
        qgrams = []

        if len(term) < self.q_param:
            return [term]

        for start in range(len(term) - self.q_param + 1):
            end = start + self.q_param
            qgrams.append(term[start:end])

        return qgrams
